--------------------------------------------------------------------------------
-- Initialization

SnowfallKeyPress = SnowfallKeyPress or {settings = {commands = {{}, {}}}};

local commandSets = {{
  {command = "^ACTIONBUTTON(%d+)$",          attributes = {{"type", "macro"}, {"macrotext",   "/click [bonusbar:5][stance] BonusActionButton%1; ActionButton%1"}}},  -- Action Buttons
  {command = "^MULTIACTIONBAR1BUTTON(%d+)$", attributes = {{"type", "macro"}, {"macrotext",   "/click MultiBarBottomLeftButton%1"                              }}},  -- BottomLeft Action Buttons
  {command = "^MULTIACTIONBAR2BUTTON(%d+)$", attributes = {{"type", "macro"}, {"macrotext",   "/click MultiBarBottomRightButton%1"                             }}},  -- BottomRight Action Buttons
  {command = "^MULTIACTIONBAR3BUTTON(%d+)$", attributes = {{"type", "macro"}, {"macrotext",   "/click MultiBarRightButton%1"                                   }}},  -- Right Action Buttons (rightmost)
  {command = "^MULTIACTIONBAR4BUTTON(%d+)$", attributes = {{"type", "macro"}, {"macrotext",   "/click MultiBarLeftButton%1"                                    }}},  -- Right ActionBar 2 Buttons (2nd from right)
  {command = "^SHAPESHIFTBUTTON(%d+)$",      attributes = {{"type", "macro"}, {"macrotext",   "/click ShapeshiftButton%1"                                      }}},  -- Special Action Buttons (shapeshift/stance)
  {command = "^BONUSACTIONBUTTON(%d+)$",     attributes = {{"type", "macro"}, {"macrotext",   "/click PetActionButton%1"                                       }}},  -- Secondary Action Buttons (pet/bonus)
  {command = "^MULTICASTRECALLBUTTON1$",     attributes = {{"type", "macro"}, {"macrotext",   "/click MultiCastRecallSpellButton"                              }}},  -- Totemic Recall
--{command = "^CLICK (.+):([^:]+)$",         attributes = {{"type", "macro"}, {"macrotext",   "/click %1 %2"                                                   }}},  -- Clicks
  {command = "^CLICK (.+):([^:]+)$",         attributes = {{"type", "click"}, {"clickbutton", "%1"                                                             }}},  -- Clicks
  {command = "^MACRO (.+)$",                 attributes = {{"type", "macro"}, {"macro",       "%1"                                                             }}},  -- Macros
--{command = "^SPELL (.+)$",                 attributes = {{"type", "spell"}, {"spell",       "%1"                                                             }}},  -- Spells
--{command = "^ITEM (.+)$",                  attributes = {{"type", "item" }, {"item",        "%1"                                                             }}},  -- Items
}};

local hook = true;

local bindingKeys = {};

local overrideFrame = CreateFrame("Frame");

local allowedTypeAttributes = {
  ["actionbar"] = true,
  ["action"] = true,
  ["pet"] = true,
  ["multispell"] = true,
  ["spell"] = true,
  ["item"] = true,
  ["macro"] = true,
  ["cancelaura"] = true,
  ["stop"] = true,
  ["target"] = true,
  ["focus"] = true,
  ["assist"] = true,
  ["maintank"] = true,
  ["mainassist"] = true
};



--------------------------------------------------------------------------------
-- Clear key binding mode so that the Blizzard key binding UI doesn't look for overrides and generate bogus messages like: "CLICK SnowfallKeyPress_Button_1:LeftButton Function is Now Unbound!"

hooksecurefunc("ShowUIPanel", function() if (KeyBindingFrame) then KeyBindingFrame.mode = nil; end end);



--------------------------------------------------------------------------------
-- Helper functions

local function enumerateFramesIterator(_, frame)
  return EnumerateFrames(frame);
end


local function addGetBindingKeys(_, ...)
  for i = 1, select("#", ...) do
    bindingKeys[select(i, ...)] = true;
  end
end


local function isSecureButton(x)
  return not not (
    type(x) == "table"
    and type(x.IsObjectType) == "function"
    and issecurevariable(x, "IsObjectType")
    and x:IsObjectType("Button")
    and select(2, x:IsProtected())
  );
end


local function accelerateKeyDoMatchingCommand(key, command, commandSets)
  local bindButtonName, bindButton;
  local attributeName, attributeValue;
  local mouseButton, harmButton, helpButton;
  local mouseType, harmType, helpType;
  local clickButtonName, clickButton;
  for _, commandSet in ipairs(commandSets) do
    for _, template in ipairs(commandSet) do
      if (string.match(command, template.command)) then
        -- make sure there are attributes. Otherwise, this key is blacklisted
        if (template.attributes) then
          clickButtonName, mouseButton = string.match(command, "^CLICK (.+):([^:]+)$");
          if (clickButtonName) then
            -- For clicks, check that the target is a SecureActionButton that isn't doing anything that could possibly rely on differentiating down/up clicks
            clickButton = _G[clickButtonName];
            if (not isSecureButton(clickButton) or clickButton:GetAttribute("", "downbutton", mouseButton)) then
              return true;
            end

            harmButton = SecureButton_GetModifiedAttribute(clickButton, "harmbutton", mouseButton);
            helpButton = SecureButton_GetModifiedAttribute(clickButton, "helpbutton", mouseButton);
            mouseType = SecureButton_GetModifiedAttribute(clickButton, "type", mouseButton);
            harmType = SecureButton_GetModifiedAttribute(clickButton, "type", harmButton);
            helpType = SecureButton_GetModifiedAttribute(clickButton, "type", helpButton);
            if (
              mouseType and not allowedTypeAttributes[mouseType]
              or harmType and not allowedTypeAttributes[harmType]
              or helpType and not allowedTypeAttributes[helpType]
            ) then
              return true;
            end
          else
            -- For non-clicks, the default mouse button is LeftButton
            mouseButton = "LeftButton";
          end

          -- make the bind button if it doesn't already exist
          bindButtonName = "SnowfallKeyPress_Button_" .. key;
          bindButton = _G[bindButtonName];
          if (not bindButton) then
            bindButton = CreateFrame("Button", "SnowfallKeyPress_Button_" .. key, nil, "SecureActionButtonTemplate");
            bindButton:RegisterForClicks("AnyDown");
          end

          -- apply specified attributes
          for _, attribute in ipairs(template.attributes) do
            attributeName = attribute[1];
            attributeValue = string.gsub(command, template.command, attribute[2], 1);

            if (attributeName == "clickbutton") then
              attributeValue = _G[attributeValue];
            end

            bindButton:SetAttribute(attributeName, attributeValue);
          end

          -- create a priority override
          hook = false;
          SetOverrideBindingClick(overrideFrame, true, key, bindButtonName, mouseButton);
          hook = true;
        end

        -- stop since we found a matching template
        return true;
      end
    end
  end
  return false;
end


-- Accelerate a key, which we must not currently be overriding
local function accelerateKey(key)
  local command;

  -- Make sure this key is bound to a command
  command = GetBindingAction(key, true);
  if (not command or command == "") then
    return;
  end

  return
    accelerateKeyDoMatchingCommand(key, command, SnowfallKeyPress.settings.commands)
    or accelerateKeyDoMatchingCommand(key, command, commandSets);
end



--------------------------------------------------------------------------------
-- UPDATE_BINDINGS
-- Find all keys we can. Accelerate them.

local function updateBindings(_, event)
  if (InCombatLockdown()) then
    return;
  end

  -- Remove all of our overrides so we can see other overrides
  hook = false;
  ClearOverrideBindings(overrideFrame);
  hook = true;

  -- Add all keys with bindings that appear in Options->KeyBindings
  for i = 1, GetNumBindings() do
    addGetBindingKeys(GetBinding(i));
  end

  -- Add all macros with bindings
  local numAccountMacros, numCharacterMacros = GetNumMacros();
  local numMacros = numAccountMacros + numCharacterMacros;
  local macroName;
  local macroCount = 0;
  for i = 1, math.huge do
    if (macroCount == numMacros) then
      break;
    end
    macroName = GetMacroInfo(i);
    if (macroName) then
      addGetBindingKeys(nil, GetBindingKey("MACRO " .. macroName));
      macroCount = macroCount + 1;
    end
  end

  -- If spell/item binding support was added, we'd add code to find their keys here

  -- Accelerate every key we know about
  for key in pairs(bindingKeys) do
    accelerateKey(key);
  end
end

overrideFrame:SetScript("OnEvent", updateBindings);
overrideFrame:RegisterEvent("UPDATE_BINDINGS");
overrideFrame:RegisterEvent("UPDATE_MACROS");



--------------------------------------------------------------------------------
-- SetOverrideBinding*
-- Make sure this key is in our table of binding keys. Remove our override. See what the key is bound to, now. Apply a new override.

local function setOverrideBindingHook(_, _, key)
  if (not hook or InCombatLockdown()) then
    return;
  end

  bindingKeys[key] = true;

  hook = false;
  SetOverrideBinding(overrideFrame, false, key, nil);
  hook = true;

  accelerateKey(key);
end

hooksecurefunc("SetOverrideBinding", setOverrideBindingHook);
hooksecurefunc("SetOverrideBindingSpell", setOverrideBindingHook);
hooksecurefunc("SetOverrideBindingClick", setOverrideBindingHook);
hooksecurefunc("SetOverrideBindingItem", setOverrideBindingHook);
hooksecurefunc("SetOverrideBindingMacro", setOverrideBindingHook);



--------------------------------------------------------------------------------
-- ClearOverrideBindings
-- Remove all our overrides. Re-apply overrides for all key bindings (to potentially new commands).

local function clearOverrideBindingsHook()
  if (not hook or InCombatLockdown()) then
    return;
  end

  hook = false;
  ClearOverrideBindings(overrideFrame);
  hook = true;

  for key in pairs(bindingKeys) do
    accelerateKey(key);
  end
end
hooksecurefunc("ClearOverrideBindings", clearOverrideBindingsHook);
